#!/usr/bin/python
#-*- coding: utf-8 -*-

#
# Date      : 2011/09/01
# Author    : Hiroyasu OHYAMA 
# Belonged  : Ariel-Network Inc. (http://www.ariel-networks.com)
#							OSSS Lab. (http://www.osss.cs.tsukuba.ac.jp)
# 
# copyright(c) 2011 Hiroyasu OHYAMA. all right reserved.
#

import re
from Operation import Operation
from Condition import Condition

from time import time

def _set_condition(cond_str):
	cond_list = re.sub(r'\s', '', cond_str).split(',')
	ret = None

	if len(cond_list) < 1 or cond_list[0] == "":
		ret = Condition("*", "*", "*")
	elif len(cond_list) < 2:
		ret = Condition(cond_list[0], "*", "*")
	elif len(cond_list) < 3:
		ret = Condition(cond_list[0], cond_list[1], "*")
	else:
		ret = Condition(cond_list[0], cond_list[1], cond_list[2])

	return ret

class Rule:
	# status const value
	STATUS_TEMPORAL = 0x01
	STATUS_ETERNAL = 0x02
	STATUS_INVALID = 0x04

	# (default value of timeout)
	# This value is enabled when status_temporal is set at status.
	DEFAULT_TIMEOUT = 10

	__INPUT_FORMAT = "^([a-zA-Z]+)\s+\((.*?)\)\s+\((.*?)\).*$"

	def __init__(self, line, status=0, timeout=DEFAULT_TIMEOUT):
		self.status = status
		self.timeout = time() + timeout

		if line != None:
			matcher = re.compile(self.__INPUT_FORMAT).search(line)
	
			self.operation = Operation(matcher.group(1))
			self.src_cond = _set_condition(matcher.group(2))
			self.dst_cond = _set_condition(matcher.group(3))

	# modifing methods of status
	def is_status(self, val):                                                                                                            
		return self.status & val                                                                                                           
	def set_status(self, val):                                                                                                           
		self.status |= val                                                                                                                 
	def del_status(self, val):                                                                                                           
		self.status &= ~val      

	# getter methods
	def get_operation(self):
		return self.operation
	def get_src_cond(self):
		return self.src_cond
	def get_dst_cond(self):
		return self.dst_cond
