#-*- coding: utf-8 -*-

from nox.lib.core	 import *

from nox.lib.packet.arp						import arp
from nox.lib.packet.tcp						import tcp
from nox.lib.packet.ipv4					import ipv4
from nox.lib.packet.ethernet			import ethernet
from nox.lib.packet.packet_utils	import mac_to_str, mac_to_int, ip_to_str

from twisted.python import log

import logging
from time import time
from socket import htons
from struct import unpack

# for rule object
from Rule.Rule import *

logger = logging.getLogger('nox.coreapps.examples.my_firewall')

# Global my_firewall instance 
inst = None

# Rule list
rule_list = None

# Timeout for cached MAC entries
CACHE_TIMEOUT = 5

# --
# Given a packet, learn the source and peg to a switch/inport 
# --
def do_l2_learning(dpid, inport, packet):
	global inst 

	# learn MAC on incoming port
	srcaddr = packet.src.tostring()
	if ord(srcaddr[0]) & 1:
		return
	if inst.st[dpid].has_key(srcaddr):
		dst = inst.st[dpid][srcaddr]
		if dst[0] != inport:
			log.msg('MAC has moved from '+str(dst)+'to'+str(inport), system='my_firewall')
		else:
			return
	else:
		log.msg('learned MAC '+mac_to_str(packet.src)+' on %d %d'% (dpid,inport), system="my_firewall")

	# learn or update timestamp of entry
	inst.st[dpid][srcaddr] = (inport, time(), packet)

	# Replace any old entry for (switch,mac).
	mac = mac_to_int(packet.src)

def pass_operation(rule, packet, ipv4pkt, tcppkt):
	
	if tcppkt != None and tcppkt.flags == tcp.SYN:
		new_rule_line = "pass (%s,*,%d) (%s,*,%d)" % (ip_to_str(ipv4pkt.dstip), tcppkt.dstport, ip_to_str(ipv4pkt.srcip), tcppkt.srcport)

		rule_list.append(Rule(new_rule_line, Rule.STATUS_TEMPORAL))

def check_packet(dpid, inport, packet, buf, bufid):
	ret = False
	rule = None

	if isinstance(packet.next, arp):
		ret = True
	elif isinstance(packet.next, ipv4):
		ipv4pkt = packet.next
	
		if isinstance(ipv4pkt.next, tcp):
			tcppkt = ipv4pkt.next

			# check processing of firewall
			rule = _will_match_rule(packet, ipv4pkt, tcppkt)

			if rule != None and rule.get_operation().get_type() == 'pass':
				pass_operation(rule, packet, ipv4pkt, tcppkt)

		else:
			rule = _will_match_rule(packet, ipv4pkt, None)
	else:
		rule = _will_match_rule(packet, None, None)

	if rule != None:
		dstaddr = packet.dst.tostring()
		ret = True

		if not ord(dstaddr[0]) & 1 and inst.st[dpid].has_key(dstaddr):
			prt = inst.st[dpid][dstaddr]
			if  prt[0] != inport:
				# We know the outport, set up a flow
				log.msg('installing flow for ' + str(packet), system="my_firewall")
				flow = extract_flow(packet)
				flow[core.IN_PORT] = inport

				# apply for firewall rule
				_set_firewall_filter(flow, rule);

				actions = [[openflow.OFPAT_OUTPUT, [0, prt[0]]]]
				inst.install_datapath_flow(dpid, flow, CACHE_TIMEOUT, 
									   openflow.OFP_FLOW_PERMANENT, actions,
									   bufid, openflow.OFP_DEFAULT_PRIORITY,
									   inport, buf)

				ret = False
	
	return ret

# --
# If we've learned the destination MAC set up a flow and
# send only out of its inport.  Else, flood.
# --
def forward_l2_packet(dpid, inport, packet, buf, bufid):	
	actions = None

	if check_packet(dpid, inport, packet, buf, bufid):
		actions = openflow.OFPP_FLOOD

	# do send packet using controller order
	if actions != None:
		inst.send_openflow(dpid, bufid, buf, actions, inport)
		
# --
# Responsible for timing out cache entries.
# Is called every 1 second.
# --
def timer_callback():
	global inst

	curtime  = time()
	for dpid in inst.st.keys():
		for entry in inst.st[dpid].keys():
			if (curtime - inst.st[dpid][entry][1]) > CACHE_TIMEOUT:
				log.msg('timing out entry'+mac_to_str(entry)+str(inst.st[dpid][entry])+' on switch %x' % dpid, system='my_firewall')
				inst.st[dpid].pop(entry)

	inst.post_callback(1, timer_callback)
	return True

def datapath_leave_callback(dpid):
	logger.info('Switch %x has left the network' % dpid)
	if inst.st.has_key(dpid):
		del inst.st[dpid]

def datapath_join_callback(dpid, stats):
	logger.info('Switch %x has joined the network' % dpid)

# --
# Packet entry method.
# Drop LLDP packets (or we get confused) and attempt learning and
# forwarding
# --
def packet_in_callback(dpid, inport, reason, len, bufid, packet):

	if not packet.parsed:
		log.msg('Ignoring incomplete packet',system='my_firewall')
		
	if not inst.st.has_key(dpid):
		log.msg('registering new switch %x' % dpid,system='my_firewall')
		inst.st[dpid] = {}

	# don't forward lldp packets	
	if packet.type == ethernet.LLDP_TYPE:
		return CONTINUE

	# learn MAC on incoming port
	do_l2_learning(dpid, inport, packet)

	forward_l2_packet(dpid, inport, packet, packet.arr, bufid)

	return CONTINUE

# construct firewall rule
def _init_rule():
	global rule_list

	rule_list = []

	f = open('/etc/firewall.conf', 'r');
	while True:
		line = f.readline()[:-1]
		if not line:
			break
	
		rule_list.append(Rule(line))
	f.close();

# this routine checks curren packet matches existed rule, or not
def _will_match_rule(packet, ipv4pkt, tcppkt):
	ret_rule = None

	for rule in rule_list:
		cond = True

		# get every parameter
		srcip = rule.get_src_cond().get_ipaddr()
		srcmac = rule.get_src_cond().get_macaddr()
		srcport = rule.get_src_cond().get_port()
		dstip = rule.get_dst_cond().get_ipaddr()
		dstmac = rule.get_dst_cond().get_macaddr()
		dstport = rule.get_dst_cond().get_port()

		cond = cond and (srcip == '*' or (ipv4pkt != None and srcip == ip_to_str(ipv4pkt.srcip)))
		cond = cond and (srcmac == '*' or (packet != None and srcmac == mac_to_str(packet.src)))
		cond = cond and (srcport == '*' or (tcppkt != None and srcport == str(tcppkt.srcport)))
		cond = cond and (dstip == '*' or (ipv4pkt != None and dstip == ip_to_str(ipv4pkt.dstip)))
		cond = cond and (dstmac == '*' or (packet != None and dstmac == mac_to_str(packet.dst)))
		cond = cond and (dstport == '*' or (tcppkt != None and dstport == str(tcppkt.dstport)))

		if cond:
			ret_rule = rule
			break

	return ret_rule

def _set_firewall_filter(attr, rule):
	attr[core.DL_SRC] = rule.get_src_cond().get_macaddr() != '*' and rule.get_src_cond().get_macaddr() or attr[core.DL_SRC]
	attr[core.DL_DST] = rule.get_dst_cond().get_macaddr() != '*' and rule.get_dst_cond().get_macaddr() or attr[core.DL_DST]
	attr[core.NW_SRC] = rule.get_dst_cond().get_ipaddr() != '*' and rule.get_dst_cond().get_ipaddr() or attr[core.NW_SRC]
	attr[core.NW_DST] = rule.get_dst_cond().get_ipaddr() != '*' and rule.get_dst_cond().get_ipaddr() or attr[core.NW_DST]
	attr[core.TP_SRC] = rule.get_dst_cond().get_port() != '*' and int(rule.get_dst_cond().get_port()) or attr[core.TP_SRC]
	attr[core.TP_DST] = rule.get_dst_cond().get_port() != '*' and int(rule.get_dst_cond().get_port()) or attr[core.TP_DST]

class my_firewall(Component):

	def __init__(self, ctxt):
		global inst
		Component.__init__(self, ctxt)
		self.st = {}
		self.ipmap = {}

		inst = self

		# initialize rule_list from input_rule file
		_init_rule()

	def install(self):
		inst.register_for_packet_in(packet_in_callback)
		inst.register_for_datapath_leave(datapath_leave_callback)
		inst.register_for_datapath_join(datapath_join_callback)
		inst.post_callback(1, timer_callback)

	def getInterface(self):
		return str(my_firewall)

def getFactory():
	class Factory:
		def instance(self, ctxt):
			return my_firewall(ctxt)

	return Factory()
